from itertools import repeat
from typing import Callable, Iterable, Tuple, TypeVar

T = TypeVar('T')
T0 = TypeVar('T0')
T1 = TypeVar('T1')


def zip_supersequence(supersequence: Iterable[T0], subsequence: Iterable[T1],
                      superseq_key: Callable[[T0], T] = None, subseq_key: Callable[[T1], T] = None,
                      subseq_default: T1 = None) -> Iterable[Tuple[T0, T1]]:
    """
    zip a supersequence and subsequence together, coupling supersequence elements without a subsequence element with a
     default value.
    Args:
        supersequence: The supersequence iterable
        subsequence: The subsequence iterable
        superseq_key: The supersequence key, if provided, equivalence to subsequence elements is made with this key
         function.
        subseq_key: The subsequence key, if provided, equivalence to supersequence elements is made with this key
         function.
        subseq_default: The value to couple with supersequence elements that have no matching subsequence elements.

    Returns:
        A generator of tuples, each member of which is a 2-tuple, of a supersequence member and either an equivalent
         subsequence member or subseq_default.

    Notes:
        The elements match eagerly

    """
    superseq_key_altered = superseq_key or (lambda x: x)  # type: ignore
    subseq_key_altered = subseq_key or (lambda x: x)  # type: ignore

    sub_iter = iter(subsequence)
    super_iter = iter(supersequence)
    try:
        next_expected_sub = next(sub_iter)
    except StopIteration:
        pass
    else:
        next_expected_sub_key = subseq_key_altered(next_expected_sub)

        for super_element in super_iter:
            super_key = superseq_key_altered(super_element)
            if super_key == next_expected_sub_key:
                yield super_element, next_expected_sub
                try:
                    next_expected_sub = next(sub_iter)
                except StopIteration:
                    break
                next_expected_sub_key = subseq_key_altered(next_expected_sub)
            else:
                yield super_element, subseq_default
        else:
            raise AssertionError('not a supersequence')

    yield from zip(super_iter, repeat(subseq_default))



from heksher.clients.async_client import AsyncHeksherClient
from heksher.clients.subclasses import TRACK_ALL
from heksher.clients.thread_client import ThreadHeksherClient
from heksher.setting import Setting

__all__ = ['AsyncHeksherClient', 'TRACK_ALL', 'ThreadHeksherClient', 'Setting']



from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from operator import attrgetter
from typing import Any, Callable, Generic, Iterable, List, Mapping, Optional, Sequence, Tuple, TypeVar, Union
from weakref import ref

from ordered_set import OrderedSet

import heksher.main_client
from heksher.clients.util import NO_DEFAULT, RuleBranch, collate_rules
from heksher.exceptions import NoMatchError
from heksher.setting_type import setting_type

logger = getLogger(__name__)

T = TypeVar('T')


class Setting(Generic[T]):
    """
    A setting object, that stores a ruleset and can be updated by heksher clients
    """

    def __init__(self, name: str, type, configurable_features: Sequence[str],
                 default_value: T = NO_DEFAULT,  # type: ignore
                 metadata: Optional[Mapping[str, Any]] = None):
        """
        Args:
            name: The name of the setting.
            type: The type of the setting, either a primitive type (int, str, bool, float), an enum class, a FlagInt
             class, or a generic alias of List or Dict.
            configurable_features: The configurable features of the setting.
            default_value: The default value of the setting, this value will be returned if no rules match the current
             context.
            metadata: Additional metadata of the setting.
        Notes:
            Creating a setting automatically registers it to be declared at the main heksher client.
        """
        self.name = name
        self.type = setting_type(type)
        self.configurable_features = OrderedSet(configurable_features)
        self.default_value = default_value
        self.metadata = metadata or {}

        self.last_ruleset: Optional[RuleSet] = None

        self._validators: List[Callable[[T, Sequence[Tuple[str, str]], Setting], T]] = []

        heksher.main_client.Main.add_settings((self,))

    def add_validator(self, validator: Callable[[T, Sequence[Tuple[str, str]], Setting], T]) \
            -> Callable[[T, Sequence[Tuple[str, str]], Setting], T]:
        """
        Add a validator to be called when the setting's value is updated.
        If multiple callbacks are added, they are called in the order they are added.
        Args:
            validator: the validator to be added.
            A validator must have three positional arguments: value, rule and Setting.
            The return value of the validator will be the new value of the rule.

        Returns:
            The validator given, so the method can be used as a decorator
        """
        self._validators.append(validator)
        return validator

    def get(self, **contexts) -> T:
        """
        Get the value of the setting for the context features.
        Args:
            **contexts: The context features along with their values, to use to resolve the ruleset. Any missing context
             features must be filled in by the client.

        Returns:
            The value for the rule matching the context, or the default value.

        Raises:
            NoMatchError if no rules matched and a default value is not defined.
        """
        redundant_keys = contexts.keys() - self.configurable_features
        if redundant_keys:
            raise ValueError(f'the following keys are not configurable: {redundant_keys}')

        if self.last_ruleset:
            try:
                from_rules = self.last_ruleset.resolve(contexts, self)
            except NoMatchError:
                from_rules = NO_DEFAULT
        else:
            logger.warning('the value of setting was never retrieved from service', extra={'setting': self.name})
            from_rules = NO_DEFAULT

        if from_rules is NO_DEFAULT:
            if self.default_value is NO_DEFAULT:
                raise NoMatchError(self.name)
            return self.default_value
        return from_rules

    def update(self, client, context_features: Sequence[str], rules: Iterable[Tuple[Sequence[Tuple[str, str]], T]]):
        """
        Update a setting's rules from a client

        Args:
            client: The client updating the setting.
            context_features: The context features the root was collated by.
            rules: An iterable of rules to collate.
        """
        if self.last_ruleset:
            last_client = self.last_ruleset.client()
            if last_client and last_client is not client:
                logger.warning('setting received rule set from multiple clients',
                               extra={'setting': self.name, 'new_client': client, 'last_client': last_client})

        def validate(rule: Sequence[Tuple[str, str]], value: T) -> T:
            for validator in self._validators:
                value = validator(value, rule, self)
            return value

        updated_rules = [(rule, validate(rule, value)) for (rule, value) in rules]
        root = collate_rules(context_features, updated_rules)
        self.last_ruleset = RuleSet(ref(client), context_features, root)


@dataclass(frozen=True)
class RuleMatch(Generic[T]):
    """
    An internal structure for resolution, representing a value belonging to a rule that matched the
     current namespace
    """
    value: T
    """
    The value of the matched rule
    """
    exact_match_depth: int
    """
    The index of the last context feature that had an exact match condition within the rule
    """


@dataclass(frozen=True)
class RuleSet(Generic[T]):
    """
    A complete set of rules, resolvable through a namespace
    """
    client: ref
    """
    A weakref to the client that supplied the ruleset
    """
    context_features: Sequence[str]
    """
    The context features the root rulebranch was collated against
    """
    root: RuleBranch[T]
    """
    The root rulebranch
    """

    def resolve(self, context_namespace: Mapping[str, str], setting: Setting):
        """
        Args:
            context_namespace: A namespace of context features and their values. If the client is available, its
             context_namespace method is used beforehand.

        Returns:
            The value of the deepest-matched rule

        Raises:
            NoMatchError, if no rules matched the namespace
        """
        client = self.client()
        if client:
            context_namespace = client.context_namespace(context_namespace)

        def _resolve(current: RuleBranch[T], depth: int = 0, exact_match_depth: int = -1) \
                -> Union[RuleMatch, bool]:  # my kingdom for a Literal!
            """
            Args:
                current: The branch being resolved
                depth: The depth of the branch (in relation to the root)
                exact_match_depth: The depth of the latest exact-match in the path from root to current, where -1
                 indicates that no exact match occurred.

            Returns:
                Either False for no matches within the branch, or a RuleMatch with the maximal branch.

            """
            if depth == len(self.context_features):
                # leaf node
                return RuleMatch(current, exact_match_depth)

            assert isinstance(current, Mapping)

            feature = self.context_features[depth]
            if feature in setting.configurable_features:
                feature_value = context_namespace.get(feature)
                if feature_value is None:
                    raise RuntimeError(f'configurable context feature {feature} is missing from both the default'
                                       ' namespace and arguments')

                exact = (feature_value in current) and _resolve(current[feature_value], depth + 1, depth)
            else:
                exact = None

            wildcard = (None in current) and _resolve(current[None], depth + 1, exact_match_depth)

            if exact and wildcard:
                # in case both wildcard and exact produced a match, we want the match with the deepest exact condition,
                # with tie-break advantage to the exact
                return max(exact, wildcard, key=attrgetter('exact_match_depth'))
            return exact or wildcard

        ret = _resolve(self.root)
        if not ret:
            raise NoMatchError
        assert isinstance(ret, RuleMatch)
        return ret.value



from __future__ import annotations

from asyncio import CancelledError, Event, Lock, Queue, Task, TimeoutError, create_task, wait_for
from contextvars import ContextVar
from datetime import datetime
from logging import getLogger
from typing import Any, Dict, NoReturn, Optional, Sequence, TypeVar, Union

import orjson
from httpx import AsyncClient, HTTPError
from ordered_set import OrderedSet

from heksher.clients.subclasses import AsyncContextManagerMixin, ContextFeaturesMixin, V1APIClient
from heksher.clients.util import SettingsOutput
from heksher.setting import NO_DEFAULT, Setting

logger = getLogger(__name__)

T = TypeVar('T')
content_header = {"Content-type": "application/json"}

__all__ = ['AsyncHeksherClient']


class AsyncHeksherClient(V1APIClient, ContextFeaturesMixin, AsyncContextManagerMixin):
    """
    An asynchronous heksher client, using heksher's V1 HTTP API
    """

    def __init__(self, service_url: str, update_interval: float, context_features: Sequence[str], *,
                 http_client_args: Dict[str, Any] = None):
        """
        Args:
            service_url: The HTTP url to the Heksher server.
            update_interval: The interval to wait between any two regular update calls, in seconds.
            context_features: The context features to expect in the Heksher server.
            http_client_args: Forwarded as kwargs to httpx.AsyncClient constructor.
        """
        super().__init__(context_features)

        http_client_args = http_client_args or {}

        self._service_url = service_url
        self._update_interval = update_interval
        self._http_client = AsyncClient(base_url=service_url, **http_client_args)
        self._undeclared: Queue[Setting] = Queue()

        self._declaration_task: Optional[Task[NoReturn]] = None
        self._update_task: Optional[Task[NoReturn]] = None

        self._last_cache_time: Optional[datetime] = None

        self.modification_lock = Lock()
        """
        A lock that is acquired whenever setting values are updated. To ensure that no modifications are made to
        settings, acquire this lock.
        """

        self._update_event = Event()
        """This event marks that an update occurred"""
        self._manual_update = Event()
        """This event is waited on by the update loop (with a timeout), set it to instantly begin an update"""

    async def _declaration_loop(self) -> NoReturn:
        """
        The method for the task that continuously declares new settings.
        """

        async def declare_setting(setting):
            declaration_data = {
                'name': setting.name,
                'configurable_features': list(setting.configurable_features),
                'type': setting.type.heksher_string(),
                'metadata': setting.metadata,
            }
            if setting.default_value is not NO_DEFAULT:
                declaration_data['default_value'] = setting.default_value
            response = await self._http_client.put('api/v1/settings/declare', content=orjson.dumps(declaration_data),
                                                   headers=content_header)
            self._handle_declaration_response(setting, response)

        while True:
            setting = await self._undeclared.get()
            try:
                await declare_setting(setting)
            except CancelledError:
                # in 3.7, cancelled is a normal exception
                raise
            except Exception:
                logger.exception('setting declaration failed',
                                 extra={'setting': setting.name})
            finally:
                self._undeclared.task_done()

    async def _update_loop(self) -> NoReturn:
        """
        The method for the task that continuously updates declared settings.
        """

        async def update():
            logger.debug('heksher reload started')
            data = {
                'setting_names': list(self._tracked_settings.keys()),
                'context_features_options': self._context_feature_options(),
                'include_metadata': False,
            }
            if self._last_cache_time:
                data['cache_time'] = self._last_cache_time.isoformat()
            new_cache_time = datetime.utcnow()

            response = await self._http_client.post('/api/v1/rules/query', content=orjson.dumps(data),
                                                    headers=content_header)
            response.raise_for_status()

            updated_settings = response.json()['rules']
            async with self.modification_lock:
                self._update_settings_from_query(updated_settings)
            self._last_cache_time = new_cache_time
            logger.info('heksher reload done', extra={'updated_settings': list(updated_settings.keys())})

        while True:
            try:
                await update()
            except CancelledError:
                # in 3.7, cancelled is a normal exception
                raise
            except Exception:
                logger.exception('error during heksher update')
            finally:
                self._update_event.set()

            try:
                self._manual_update.clear()
                # we only want to pass immediately if the event was passed while we are waiting
                await wait_for(self._manual_update.wait(), self._update_interval)
            except TimeoutError:
                pass

    async def set_as_main(self):
        await super().set_as_main()

        # check that we're dealing with the right context features
        try:
            response = await self._http_client.get('/api/v1/context_features')
            response.raise_for_status()
        except HTTPError:
            logger.exception('failure to get context_features from heksher service',
                             extra={'service_url': self._service_url})
        else:
            features_in_service = response.json()['context_features']
            if features_in_service != self._context_features:
                logger.warning('context feature mismatch', extra={
                    'features_in_service': features_in_service,
                    'features_in_client': self._context_features
                })
                # we let the service decide the features to avoid future conflict
                self._context_features = OrderedSet(features_in_service)

        self._declaration_task = create_task(self._declaration_loop())
        # important to only start the update thread once all pending settings are declared, otherwise we may have
        # stale settings
        await self._undeclared.join()
        self._update_task = create_task(self._update_loop())
        await self.reload()

    async def reload(self):
        """
        Block until all the tracked settings are up to date
        """
        await self._undeclared.join()
        self._update_event.clear()
        self._manual_update.set()
        await self._update_event.wait()

    async def close(self):
        await super().close()
        if self._update_task:
            self._update_task.cancel()
            try:
                await wait_for(self._update_task, 0.01)
            except CancelledError:
                pass
            self._update_task = None
        if self._declaration_task:
            self._declaration_task.cancel()
            try:
                await wait_for(self._declaration_task, 0.01)
            except CancelledError:
                pass
            self._declaration_task = None
        await self._http_client.aclose()

    def set_defaults(self, **kwargs: Union[str, ContextVar[str]]):
        redundant_keys = kwargs.keys() - set(self._context_features)
        if redundant_keys:
            logger.warning('context features are not specified in the client', extra={
                'redundant_keys': redundant_keys
            })
        super().set_defaults(**kwargs)

    async def ping(self) -> None:
        """
        Check the health of the heksher server
        Raises:
            httpx.HTTPError, if an error occurs
        """
        response = await self._http_client.get('/api/health')
        response.raise_for_status()

    async def get_settings(self) -> Dict:
        """
        List all the settings in the service
        """
        response = await self._http_client.get('/api/v1/settings', params=orjson.dumps(
            {'include_additional_data': True}))
        response.raise_for_status()
        settings = SettingsOutput.parse_obj(response.json()).to_settings_data()
        return settings



from __future__ import annotations

from dataclasses import dataclass
from itertools import groupby
from typing import Collection, ContextManager, Generic, Mapping, Optional, TypeVar, Union
from unittest.mock import MagicMock

try:
    from unittest.mock import AsyncMock  # type: ignore[attr-defined]
except ImportError:
    # asyncmock only available for python 3.8 and up, for earlier versions, we use the backport
    from mock import AsyncMock  # type: ignore[attr-defined]

import heksher.main_client
from heksher.clients.subclasses import AsyncContextManagerMixin, ContextFeaturesMixin, ContextManagerMixin
from heksher.setting import Setting

T = TypeVar('T')

__all__ = ['Rule', 'SyncStubHeksherClient', 'AsyncStubHeksherClient']


@dataclass
class Rule(Generic[T]):
    """
    A rule for stub clients, to emulate setting rules
    """
    match_conditions: Mapping[str, Optional[str]]
    """
    The match conditions for the rule, with "None" as a wildcard.
    Notes:
        in StubClients, when multiple rules are given, they must have the exact same keys, in exactly the same order.
        These keys will be interpreted as the setting's context features.
    """
    value: T
    """
    The value of a setting that matches the conditions
    """


class StubClient(ContextFeaturesMixin):
    """
    An abstract stub heksher client, usable for testing.
    """

    class _Patch(ContextManager):
        """
        A context manager for a stubbed setting
        """

        def __init__(self, setting: Setting):
            self.setting = setting
            self.previous = setting.last_ruleset

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.setting.last_ruleset = self.previous
            return None

    def add_settings(self, settings):
        pass

    def track_contexts(self, **context_values: Union[str, Collection[str]]):
        pass

    def patch(self, setting: Setting[T], value: Union[T, Collection[Rule[T]]]) -> ContextManager:
        """
        Stub the value of a particular setting.

        Args:
            setting: The setting whose value to set.
            value: The value of the setting, or a collection of rules to be collated.

        Returns:
            A context manager that will reset the value of the setting upon exit.
        """
        ret = self._Patch(setting)  # initialize the patch first so it can know the previous value
        # check for rule collection
        if isinstance(value, Collection) and isinstance(next(iter(value), None), Rule):
            # we need to make sure all the rules have exactly the same context_features
            grouped = groupby(value, lambda rule: tuple(rule.match_conditions.keys()))
            first, _ = next(grouped)
            try:
                second, _ = next(grouped)
            except StopIteration:
                pass
            else:
                raise RuntimeError('StubClient.patch with a rule list requires that all the rules have the exact'
                                   f' same features (got {first} and {second})')
            rules = [
                ([(k, v) for k, v in rule.match_conditions.items() if v is not None], rule.value)
                for rule in value
            ]
            setting.update(self, first, rules)
        else:
            setting.update(self, (), [((), value)])  # type: ignore

        return ret

    def _set_as_main(self):
        heksher.main_client.Main = self


class AsyncStubHeksherClient(StubClient, AsyncContextManagerMixin):
    """
    An asynchronous heksher client. Compatible with AsyncHeksherClient.
    """
    close = None

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.reload = AsyncMock()
        self.close = AsyncMock()
        self.ping = AsyncMock()

    async def set_as_main(self):
        super()._set_as_main()


class SyncStubHeksherClient(StubClient, ContextManagerMixin):
    """
    A synchronous heksher client. Compatible with ThreadHeksherClient.
    """
    close = None

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.reload = MagicMock()
        self.close = MagicMock()
        self.ping = MagicMock()

    def set_as_main(self):
        super()._set_as_main()



from __future__ import annotations

import collections.abc
from abc import ABC, abstractmethod
from enum import Enum, IntFlag
from logging import getLogger
from types import MappingProxyType
from typing import Any, Optional, Tuple, Type

import orjson

try:
    from types import GenericAlias  # type: ignore[attr-defined]
except ImportError:
    GenericAlias = None  # type: ignore[misc]

try:
    from typing import get_args, get_origin  # type: ignore[attr-defined]
except ImportError:
    # functions only available for 3.8 and up
    def get_args(tp: Any) -> Tuple[Any, ...]:
        return getattr(tp, '__args__', ())

    def get_origin(tp: Any) -> Optional[Any]:
        return getattr(tp, '__origin__', None)

logger = getLogger(__name__)


class SettingType(ABC):
    """
    Base class for setting types
    """

    @abstractmethod
    def heksher_string(self) -> str:
        """
        Returns: The type as string, as specified by the Heksher specs
        """
        pass

    @abstractmethod
    def convert(self, x):
        """
        Args:
            x: JSON-parsed value, retrieved from http api

        Returns:
            x, converted to an immutable pythonic value

        Notes:
            convert must return an immutable value
        """
        pass


class SimpleSettingType(SettingType):
    """
    A setting type for immutable primitives
    """

    def __init__(self, name):
        """
        Args:
            name: The name of the primitive in heksher specs
        """
        self.name = name

    def heksher_string(self) -> str:
        return self.name

    def convert(self, x):
        return x


class FlagsType(SettingType):
    """
    A setting type for flags, reflecting a flags of strings in heksher service.
    Notes:
        Although the heksher type is a flag of strings, the python type must be an IntFlags, where the strings are the
         member names.
    """

    def __init__(self, flags_type: Type[IntFlag]):
        """
        Args:
            flags_type: The IntFlags subclass to use as a type
        """
        self.type_ = flags_type

    def heksher_string(self) -> str:
        return 'Flags[' + ','.join(sorted(str(orjson.dumps(x), 'utf-8') for x in self.type_.__members__)) + ']'

    def convert(self, x):
        ret = self.type_(0)
        for i in x:
            if not isinstance(i, str):
                raise TypeError(f'expected string in flags, got {type(i).__name__}')
            try:
                member = self.type_[i]
            except KeyError:
                logger.error('server sent a flag value not found in the python type',
                             extra={'setting_type': self.heksher_string(), 'received value': i})
            else:
                ret |= member
        return ret


class EnumType(SettingType):
    """
    A setting type for an enum of primitive values
    """

    def __init__(self, enum_type: Type[Enum]):
        """
        Args:
            enum_type: The Enum subclass to use as a type
        """
        self.type_ = enum_type

    def heksher_string(self) -> str:
        return 'Enum[' + ','.join(sorted(str(orjson.dumps(x.value), 'utf-8') for x in self.type_)) + ']'

    def convert(self, x):
        return self.type_(x)


class GenericSequenceType(SettingType):
    """
    A setting type for a sequence type
    """

    def __init__(self, inner: SettingType):
        """
        Args:
            inner: the inner setting type of each member
        """
        self.inner = inner

    def heksher_string(self) -> str:
        return f'Sequence<{self.inner.heksher_string()}>'

    def convert(self, x):
        return tuple(self.inner.convert(i) for i in x)


class GenericMappingType(SettingType):
    """
    A setting type for a mapping type with string keys
    """

    def __init__(self, inner: SettingType):
        """
        Args:
            inner: the inner setting type of each value in the
        """
        self.inner = inner

    def heksher_string(self) -> str:
        return f'Mapping<{self.inner.heksher_string()}>'

    def convert(self, x):
        return MappingProxyType({k: self.inner.convert(v) for (k, v) in x.items()})


_simples = {
    int: SimpleSettingType('int'),
    float: SimpleSettingType('float'),
    str: SimpleSettingType('str'),
    bool: SimpleSettingType('bool'),
}


def setting_type(py_type: Any) -> SettingType:
    """
    Parse a python type to a heksher setting type
    Args:
        py_type: the python type or genetic alias to parse

    Returns:
        A SettingType respective of py_type

    """
    if isinstance(py_type, type):
        simple = _simples.get(py_type)
        if simple:
            return simple
        if issubclass(py_type, IntFlag):
            return FlagsType(py_type)
        if issubclass(py_type, Enum):
            return EnumType(py_type)
    # we can't depend on GenericAlias to act the way we expect it to, we instead use get_origin and get_args
    if get_origin(py_type) is collections.abc.Sequence:
        arg, = get_args(py_type)
        inner = setting_type(arg)
        return GenericSequenceType(inner)
    if get_origin(py_type) is collections.abc.Mapping:
        key_type, value_type = get_args(py_type)
        if key_type is not str:
            raise TypeError('the key for mapping setting types must always be str')
        inner = setting_type(value_type)
        return GenericMappingType(inner)
    raise RuntimeError(f'could not convert python value {py_type} to setting type')



from __future__ import annotations

import asyncio
import queue
from abc import ABC, abstractmethod
from collections import defaultdict
from contextvars import ContextVar
from logging import getLogger
from typing import (
    AsyncContextManager, Collection, ContextManager, Dict, Iterable, Mapping, MutableMapping, Sequence, Set, Tuple,
    TypeVar, Union
)
from weakref import WeakValueDictionary

from httpx import Response
from ordered_set import OrderedSet

import heksher.main_client
from heksher.clients.util import collate_rules
from heksher.heksher_client import BaseHeksherClient, TemporaryClient
from heksher.setting import RuleBranch, Setting

logger = getLogger(__name__)

T = TypeVar('T')

TRACK_ALL = '*'


class V1APIClient(BaseHeksherClient, ABC):
    """
    A client base class with shared logic for heksher's v1 HTTP API.
    """
    _undeclared: Union[asyncio.Queue, queue.Queue]
    _context_features: OrderedSet[str]

    def __init__(self, context_features: Sequence[str]):
        """
        Args:
            context_features: The context features to store as default in the client
        """
        super().__init__()
        self._context_features: OrderedSet[str] = OrderedSet(context_features)

        self._tracked_context_options: Dict[str, Union[Set[str], str]] = defaultdict(set)
        # the tracked options can also include the sentinel value TRACK_ALL
        # value will always be a set or TRACK_ALL, Literal is not supported in python 3.7
        self._tracked_settings: MutableMapping[str, Setting] = WeakValueDictionary()

    def collate_rules(self, rules: Iterable[Tuple[Sequence[Tuple[str, str]], T]]) -> RuleBranch[T]:
        """
        Collate a list of rules, according to the client's context features

        Args:
            rules: an iterable of rules

        Returns:
            The root rule branch

        """
        return collate_rules(self._context_features, rules)

    def add_settings(self, settings: Iterable[Setting]):
        for s in settings:
            self._undeclared.put_nowait(s)

    def track_contexts(self, **context_values: Union[str, Collection[str]]):
        """
        Register the context feature options to be tracked by the client. An option must be tracked for any rules
         pertaining to it to be used.

        Args:
            **context_values: context features, mapped to either a single possible value or a collection.
        """
        redundant_keys = context_values.keys() - self._context_features
        if redundant_keys:
            logger.warning('context features are not specified in server', extra={
                'redundant_keys': redundant_keys
            })
        for k, v in context_values.items():
            if v == TRACK_ALL:
                if self._tracked_context_options.get(k) is not None:
                    raise RuntimeError("cannot set TRACK_ALL to a feature that's already been used to track")
                self._tracked_context_options[k] = TRACK_ALL
                continue
            if isinstance(v, str):
                v = (v,)
            if self._tracked_context_options[k] == TRACK_ALL:
                raise RuntimeError("cannot track a specific value after the feature's been set to TRACK_ALL")
            self._tracked_context_options[k].update(v)  # type: ignore

    def _context_feature_options(self):
        return {k: (TRACK_ALL if v == TRACK_ALL else list(v))
                for k, v in self._tracked_context_options.items()}

    def _handle_declaration_response(self, setting: Setting, response: Response):
        """
        Inner utility method to handle a setting declaration response
        Args:
            setting: the setting declared
            response: the http response from the service

        """
        response.raise_for_status()
        response_data = response.json()
        incomplete = response_data['incomplete']
        if incomplete:
            logger.warning('some setting entries are incomplete in declaration',
                           extra={'setting': setting.name, 'incomplete': incomplete})
        changed = response_data['changed']
        if changed:
            logger.warning('some setting entries were changed by declaration',
                           extra={'setting': setting.name, 'changed': changed})
        logger.info('setting declared', extra={'setting': setting.name})
        self._tracked_settings[setting.name] = setting

    def _update_settings_from_query(self, updated_settings: Mapping[str, dict]):
        """
        Inner utility method to handle a setting update
        Args:
            updated_settings: A mapping of settings, with updated rules from the HTTP service
        """
        for setting_name, rule_mappings in updated_settings.items():
            setting = self._tracked_settings[setting_name]
            rules = ((rm['context_features'], setting.type.convert(rm['value'])) for rm in rule_mappings)
            setting.update(self, self._context_features, rules)

    def context_namespace(self, user_namespace: Mapping[str, str]) -> Mapping[str, str]:
        redundant_keys = user_namespace.keys() - self._context_features
        if redundant_keys:
            logger.warning('context features are not specified in server', extra={
                'redundant_keys': redundant_keys
            })

        for k, v in user_namespace.items():
            if k in redundant_keys:
                # all redundant keys have already been handled
                continue
            tracked_contexts_options = self._tracked_context_options.get(k, ())
            if v not in tracked_contexts_options and tracked_contexts_options != TRACK_ALL:
                logger.warning('context feature value is not tracked by client',
                               extra={'context_feature': k, 'context_feature_value': v})
        return super().context_namespace(user_namespace)

    def _set_as_main(self):
        """
        Transfer "main-ness" to self, either from a temporary client or another v1 client
        """
        previous_main: BaseHeksherClient = heksher.main_client.Main  # for readability of this function
        if isinstance(previous_main, TemporaryClient):
            self.add_settings(previous_main.undeclared)
        elif isinstance(previous_main, V1APIClient):
            # know when you've fucked up
            logger.warning("switching main heksher clients! this is NOT recommended! (unless this is a test)")
            if not previous_main._undeclared.empty():
                raise RuntimeError("previous main heksher client still has unprocessed declarations, "
                                   "did you forget to close it?")
            if self._context_features != previous_main._context_features:
                # now you've really fucked up
                raise RuntimeError(f"new main heksher client has different contexts, "
                                   f"previous: {previous_main._context_features}, new: {self._context_features}")
            # if we are using the same contexts, we can safely add the same settings to this client
            # no need to redeclare them, so we just track them
            for setting_name, setting in previous_main._tracked_settings.items():
                if setting_name in self._tracked_settings:
                    continue
                self._tracked_settings[setting.name] = setting
            if self._tracked_context_options != previous_main._tracked_context_options:
                # this won't cause errors, but it will cause some settings to have the wrong values, be warned
                logger.warning("new main heksher client tracks different context options",
                               extra={"previous_options": previous_main._tracked_context_options,
                                      "new_options": self._tracked_context_options})
        else:
            raise TypeError(f'cannot change main client from type '
                            f'{type(previous_main).__name__} to type {type(self).__name__}')
        heksher.main_client.Main = self


class ContextFeaturesMixin(BaseHeksherClient, ABC):
    """
    A mixin class to handle default context feature values, either from contextvars or string constants.
    """
    _context_features: Collection[str]

    def __init__(self):
        super().__init__()
        self._const_context_features: MutableMapping[str, str] = {}
        self._contextvar_context_features: MutableMapping[str, ContextVar[str]] = {}

    def set_defaults(self, **kwargs: Union[str, ContextVar[str]]):
        existing_keys = kwargs.keys() & (self._const_context_features.keys() | self._contextvar_context_features.keys())
        if existing_keys:
            raise ValueError(f'defaults for context features {existing_keys} have already been added')

        for k, v in kwargs.items():
            if isinstance(v, str):
                self._const_context_features[k] = v
            else:
                self._contextvar_context_features[k] = v

    def context_namespace(self, user_namespace: Mapping[str, str]) -> Mapping[str, str]:
        ret = dict(self._const_context_features)
        ret.update(user_namespace)

        for k, cv in self._contextvar_context_features.items():
            if k in ret:
                # skip the context value since the ns already provided a value
                continue

            try:
                value = cv.get()
            except LookupError:
                continue
            ret[k] = value
        return ret


class ContextManagerMixin(BaseHeksherClient, ContextManager):
    """
    Mixin class to treat a client as a synchronous context manager
    """

    @abstractmethod
    def set_as_main(self):
        self._set_as_main()

    @abstractmethod
    def close(self):
        pass

    def __enter__(self):
        self.set_as_main()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncContextManagerMixin(BaseHeksherClient, AsyncContextManager):
    """
    Mixin class to treat a client as an asynchronous context manager
    """

    @abstractmethod
    async def set_as_main(self):
        self._set_as_main()

    @abstractmethod
    async def close(self):
        pass

    async def __aenter__(self):
        await self.set_as_main()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



from __future__ import annotations

from contextvars import ContextVar
from datetime import datetime
from logging import getLogger
from queue import Empty, Queue
from threading import Event, Lock, Thread
from typing import Any, Dict, Optional, Sequence, TypeVar, Union

import orjson
from httpx import Client, HTTPError
from ordered_set import OrderedSet

from heksher.clients.subclasses import ContextFeaturesMixin, ContextManagerMixin, V1APIClient
from heksher.clients.util import SettingsOutput
from heksher.setting import NO_DEFAULT, Setting

logger = getLogger(__name__)

T = TypeVar('T')
content_header = {"Content-type": "application/json"}

__all__ = ['ThreadHeksherClient']


class ThreadHeksherClient(V1APIClient, ContextFeaturesMixin, ContextManagerMixin):
    """
    A synchronous heksher client, using heksher's V1 HTTP API
    """
    declaration_loop_interval = 1
    """
    The timeout when waiting for new settings to declare. This is the maximum blocking time of the declare
     thread closing.
    """

    def __init__(self, service_url: str, update_interval: float, context_features: Sequence[str], *,
                 http_client_args: Dict[str, Any] = None):
        """
        Args:
            service_url: The HTTP url to the Heksher server.
            update_interval: The interval to wait between any two regular update calls, in seconds.
            context_features: The context features to expect in the Heksher server.
            http_client_args: Forwarded as kwargs to httpx.AsyncClient constructor.
        """
        super().__init__(context_features)

        self._service_url = service_url
        self._update_interval = update_interval
        self._http_client_args = http_client_args or {}
        self._undeclared: Queue[Setting] = Queue()

        self._declaration_thread: Optional[Thread] = None
        self._update_thread: Optional[Thread] = None

        self._last_cache_time: Optional[datetime] = None

        self.modification_lock = Lock()
        """
        A lock that is acquired whenever setting values are updated. To ensure that no modifications are made to
        settings, acquire this lock.
        """

        self._update_event = Event()
        """This event marks that an update occurred"""
        self._manual_update = Event()
        """This event is waited on by the update loop on a timeout, set it to true to instantly begin an update"""

        self._keep_going = True

    def _http_client(self):
        """
        Create an httpx client to interface with the service.
        Notes:
            httpx clients are not thread-safe
        """
        return Client(base_url=self._service_url, **self._http_client_args)

    def _declare_loop(self):
        """
        thread target to continuously declare new settings.
        """
        http_client = self._http_client()

        def declare_setting(setting):
            declaration_data = {
                'name': setting.name,
                'configurable_features': list(setting.configurable_features),
                'type': setting.type.heksher_string(),
                'metadata': setting.metadata,
            }
            if setting.default_value is not NO_DEFAULT:
                declaration_data['default_value'] = setting.default_value
            response = http_client.put('api/v1/settings/declare', content=orjson.dumps(declaration_data),
                                       headers=content_header)
            self._handle_declaration_response(setting, response)

        while self._keep_going:
            try:
                setting = self._undeclared.get(timeout=self.declaration_loop_interval)
            except Empty:
                continue

            try:
                declare_setting(setting)
            except Exception:
                logger.exception('declaration failed', extra={'setting': setting.name})
            finally:
                self._undeclared.task_done()

    def _update_loop(self):
        """
        Thread target to continuously update declared settings.
        """

        http_client = self._http_client()

        def update():
            logger.debug('heksher reload started')
            data = {
                'setting_names': list(self._tracked_settings.keys()),
                'context_features_options': self._context_feature_options(),
                'include_metadata': False,
            }
            if self._last_cache_time:
                data['cache_time'] = self._last_cache_time.isoformat()
            new_cache_time = datetime.utcnow()

            response = http_client.post('/api/v1/rules/query', content=orjson.dumps(data),
                                        headers=content_header)
            response.raise_for_status()

            updated_settings = response.json()['rules']
            with self.modification_lock:
                self._update_settings_from_query(updated_settings)
            self._last_cache_time = new_cache_time
            logger.info('heksher reload done', extra={'updated_settings': updated_settings.keys()})

        while self._keep_going:
            try:
                update()
            except Exception:
                logger.exception('error during heksher update')
            finally:
                self._update_event.set()

            self._manual_update.clear()
            self._manual_update.wait(self._update_interval)

    def set_as_main(self):
        super().set_as_main()

        # check that we're dealing with the right context features
        try:
            response = self._http_client().get('/api/v1/context_features')
            response.raise_for_status()
        except HTTPError:
            logger.exception('failure to get context_features from heksher service',
                             extra={'service_url': self._service_url})
        else:
            features_in_service = response.json()['context_features']
            if features_in_service != self._context_features:
                logger.warning('context feature mismatch', extra={
                    'features_in_service': features_in_service,
                    'features_in_client': self._context_features
                })
                # we let the service decide the features to avoid future conflict
                self._context_features = OrderedSet(features_in_service)

        self._declaration_thread = Thread(target=self._declare_loop, daemon=True)
        self._declaration_thread.start()
        self._undeclared.join()
        # important to only start the update thread once all pending settings are declared, otherwise we may have
        # stale settings
        self._update_thread = Thread(target=self._update_loop, daemon=True)
        self._update_thread.start()
        self.reload()

    def reload(self):
        """
        Block until all the tracked settings are up to date
        """
        self._undeclared.join()
        self._update_event.clear()
        self._manual_update.set()
        self._update_event.wait()

    def close(self):
        super().close()
        self._keep_going = False
        self._manual_update.set()
        self._update_thread.join()
        self._declaration_thread.join()

    def set_defaults(self, **kwargs: Union[str, ContextVar[str]]):
        redundant_keys = kwargs.keys() - set(self._context_features)
        if redundant_keys:
            logger.warning('context features are not specified in the client', extra={
                'redundant_keys': redundant_keys
            })
        super().set_defaults(**kwargs)

    def ping(self) -> None:
        """
        Check the health of the heksher server
        Raises:
            httpx.HTTPError, if an error occurs
        """
        response = self._http_client().get('/api/health')
        response.raise_for_status()

    def get_settings(self) -> Dict:
        """
        List all the settings in the service
        """
        response = self._http_client().get('/api/v1/settings', params=orjson.dumps(
            {'include_additional_data': True}))
        response.raise_for_status()
        settings = SettingsOutput.parse_obj(response.json()).to_settings_data()
        return settings



from dataclasses import dataclass
from logging import getLogger
from operator import itemgetter
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import orjson
from pydantic import BaseModel

from heksher.util import zip_supersequence

logger = getLogger(__name__)
T = TypeVar('T')

NO_DEFAULT = object()


RuleBranch = Union[Mapping[Optional[str], 'RuleBranch[T]'], T]  # type: ignore[misc]
"""
A RuleBranch is a nested collation of rules or sub-rules, stored in a uniform-depth tree structure.
For example, the following set of rules:
{user: john} -> 100
{user: jim, trust: admin} -> 200
{user: jim} -> 50
{trust: guest, theme: dark} -> 20
{trust: guest} -> 10

Will be collated to the following rulebranch:
{
  "john": {
    None: {
      None:100
    }
  },
  "jim": {
    "admin": {
      None: 200
    },
    None: {
      None: 50
    }
  },
  None: {
    "guest": {
      "dark": 20,
      None: 10
    }
  }
}
"""


def collate_rules(keys: Sequence[str], rules: Iterable[Tuple[Sequence[Tuple[str, str]], T]]) -> RuleBranch[T]:
    """
    Gather a set of rules into a rule-branch root
    Args:
        keys: The context features to collate by
        rules: An iterable of rules to collate. Each rule is a two-tuple, first the sequence of exact-match condition
         (in hierarchical order), the second the value.

    Returns:
        The rule branch root for the rules, using the context features.

    """
    if not keys:
        # special case for no cfs, return value will be a single value
        rule_iter = iter(rules)
        try:
            conds, ret = next(rule_iter)
        except StopIteration:
            # no rules at all
            return NO_DEFAULT  # type: ignore
        assert not conds
        # we assert that there is, at most, one rule
        try:
            conds, ret = next(rule_iter)
        except StopIteration:
            pass
        else:
            assert not conds
            logger.error('rule conflict, overlapping values for context in service without context features',
                         stack_info=True)
        return ret

    root: Dict[str, Any] = {}
    for conditions, value in rules:
        # we constantly point the current node in the tree by storing its parent and the path to get there
        parent = None
        child_key = None  # root is without a path
        for cf, condition in zip_supersequence(keys, conditions, subseq_key=itemgetter(0)):
            if condition:
                # exact_match condition
                _, key = condition
            else:
                # wildcard
                key = None

            if parent is None:
                # we were pointing at the root
                parent = root
            else:
                parent = parent.setdefault(child_key, {})

            child_key = key
        if child_key in parent:
            # rule conflict
            logger.error('rule conflict, overlapping values for context', extra={'conditions': conditions},
                         stack_info=True)
        assert parent is not None
        parent[child_key] = value
    return root


def orjson_dumps(v, **kwargs):
    return str(orjson.dumps(v, **kwargs), 'utf-8')


class SingleSettingData(BaseModel):
    name: str
    configurable_features: List[str]
    type: str
    default_value: Any
    metadata: Dict[str, Any]

    def to_setting_data(self):
        return SettingData(
            name=self.name,
            configurable_features=self.configurable_features,
            type=self.type,
            default_value=self.default_value,
            metadata=self.metadata,
            )


class SettingsOutput(BaseModel):
    settings: List[SingleSettingData]

    class Config:
        json_dumps = orjson_dumps
        json_loads = orjson.loads

    def to_settings_data(self) -> Dict:
        return {setting.name: setting.to_setting_data() for setting in self.settings}


@dataclass
class SettingData:
    name: str
    configurable_features: List[str]
    type: str
    default_value: Any
    metadata: Dict[str, Any]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['heksher', 'heksher.clients']

package_data = \
{'': ['*']}

install_requires = \
['httpx',
 'ordered-set>=4.0.0,<5.0.0',
 'orjson>=3.0.0,<4.0.0',
 'pydantic>=1.0.0,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['mock>=4.0.0,<5.0.0']}

setup_kwargs = {
    'name': 'heksher',
    'version': '0.1.5',
    'description': '',
    'long_description': '# heksher SDK for python\nThis is a library for using a [heksher](https://github.com/biocatchltd/Heksher) server from within python.\nCompatible with python 3.7, 3.8, and 3.9. The library contains both an asynchronous client, as well as a thread-based\nclient. Also included are stub clients to make testing without a service simple.\n\n## Example usage\n```python\n# main.py\nfrom contextvars import ContextVar\nfrom heksher import AsyncHeksherClient, Setting\n\nuser = ContextVar(\'user\', default=\'guest\')\n\nclass App:\n    ...\n    \n    async def startup(self):\n        ...\n        \n        # initialize the client, and set it as the process\' main client\n        self.heksher_client = AsyncHeksherClient(\'http://heksher.service.url\',\n                                            update_interval=60, \n                                            context_features=[\'user\', \'trust\', \'theme\'])\n        # set certain context features to be retrieved either from string constants or\n        # context variables \n        self.heksher_client.set_defaults(user = user, theme="light")\n        await self.heksher_client.set_as_main()\n    \n    async def shutdown(self):\n        await self.heksher_client.close()        \n        ...\n\ncache_size_setting = Setting(\'cache_size\', type=int, configurable_features=[\'user\', \'trust\'], default_value=10)\ndef foo(trust: str):\n    ...\n    # should be run after App.startup is completed\n    cache_size = cache_size_setting.get(trust=trust)\n    ...\n```\nThread-based client usage is nearly identical. ',
    'author': 'Biocatch LTD',
    'author_email': 'serverteam@biocatch.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/biocatchltd/heksher-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

# avoid circular import
if TYPE_CHECKING:
    from heksher.heksher_client import BaseHeksherClient  # pragma: nocover

Main: BaseHeksherClient
"""
A central client, whose update loops run in the background, and to whom all settings declare themselves.
"""



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Mapping
from weakref import WeakSet

import heksher.main_client
from heksher.setting import Setting


class BaseHeksherClient(ABC):
    @abstractmethod
    def add_settings(self, settings: Collection[Setting]):
        """
        Register a setting to be declared.
        Args:
            settings: the setting that needs to be declared.
        """
        pass

    @abstractmethod
    def context_namespace(self, user_namespace: Mapping[str, str]) -> Mapping[str, str]:
        """
        Create a context namespace for ruleset resolution, with context features and their values.
        Args:
            user_namespace: namespace provided by the user, to be treated as overriding values.

        Returns:
            A new mapping combining ns and the client's state, to be used for ruleset resolution.

        """
        pass

    @abstractmethod
    def _set_as_main(self):
        """
        Internal method to transfer "main-ness" to self
        """
        pass


class TemporaryClient(BaseHeksherClient):
    """
    A temporary client, to hold undeclared settings until another client takes over.
    Note: Should only be used internally until a real client is set as main.
    """

    def __init__(self):
        self.undeclared: WeakSet = WeakSet()

    def add_settings(self, settings: Collection[Setting]):
        self.undeclared.update(settings)

    def context_namespace(self, user_namespace: Mapping[str, str]) -> Mapping[str, str]:
        return user_namespace

    def _set_as_main(self):
        raise RuntimeError("temporary client cannot be set as the main heksher client")


heksher.main_client.Main = TemporaryClient()
